"""The constants for aioautomower."""

from enum import IntEnum, StrEnum

API_BASE_URL = "https://api.amc.husqvarna.dev/v1"
AUTH_API_BASE_URL = "https://api.authentication.husqvarnagroup.dev/v1"
AUTH_API_TOKEN_URL = f"{AUTH_API_BASE_URL}/oauth2/token"
AUTH_API_REVOKE_URL = f"{AUTH_API_BASE_URL}/oauth2/revoke"
AUTH_HEADERS = {
    "Content-Type": "application/x-www-form-urlencoded",
    "Accept": "application/json",
}
AUTH_HEADER_FMT = "Bearer {}"
HUSQVARNA_URL = "https://developer.husqvarnagroup.cloud/"
REST_POLL_CYCLE = 300
TOKEN_URL = f"{AUTH_API_BASE_URL}/token"
USER_URL = f"{AUTH_API_BASE_URL}/users"
WS_URL = "wss://ws.openapi.husqvarna.dev/v1"


class EventTypesV2(StrEnum):
    """Websocket events from websocket V2."""

    BATTERY = "battery-event-v2"
    CALENDAR = "calendar-event-v2"
    CUTTING_HEIGHT = "cuttingHeight-event-v2"
    HEADLIGHTS = "headlights-event-v2"
    MESSAGES = "message-event-v2"
    MOWER = "mower-event-v2"
    PLANNER = "planner-event-v2"
    POSITIONS = "position-event-v2"


class DayOfWeek(IntEnum):
    """Day of the week."""

    SUNDAY = 0
    """Sunday."""

    MONDAY = 1
    """Monday."""

    TUESDAY = 2
    """Tuesday."""

    WEDNESDAY = 3
    """Wednesday."""

    THURSDAY = 4
    """Thursday."""

    FRIDAY = 5
    """Friday."""

    SATURDAY = 6
    """Saturday."""


class ProgramFrequency(IntEnum):
    """Program frequency."""

    DAILY = 0
    """A daily schedule with every days of the week."""

    WEEKLY = 1
    """A schedule that cycles every N weeks."""



"""Models for Automower Connect API - Calendar."""

from dataclasses import dataclass, field, fields
from datetime import datetime, time, timedelta
from typing import TYPE_CHECKING

from ical.iter import (
    MergedIterable,
    SortableItem,
)
from mashumaro import DataClassDictMixin, field_options
from mashumaro.config import BaseConfig
from mashumaro.types import SerializationStrategy

from aioautomower import tz_util
from aioautomower.const import DayOfWeek, ProgramFrequency
from aioautomower.timeline import ProgramEvent, ProgramTimeline, create_recurrence

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ical.timespan import Timespan

WEEKDAYS = (
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
)

WEEKDAYS_TO_ICAL = {
    "sunday": DayOfWeek.SUNDAY,
    "monday": DayOfWeek.MONDAY,
    "tuesday": DayOfWeek.TUESDAY,
    "wednesday": DayOfWeek.WEDNESDAY,
    "thursday": DayOfWeek.THURSDAY,
    "friday": DayOfWeek.FRIDAY,
    "saturday": DayOfWeek.SATURDAY,
}


class TimeSerializationStrategy(SerializationStrategy):
    """SerializationStrategy for Recur object."""

    def serialize(self, value: time) -> int:
        """Serialize a time object to an integer representing minutes."""
        return value.hour * 60 + value.minute

    def deserialize(self, value: int) -> time:
        """Deserialize an integer to a time object."""
        hour = value // 60
        minute = value % 60
        return time(hour=hour, minute=minute)


class DurationSerializationStrategy(SerializationStrategy):
    """SerializationStrategy for timedelta object."""

    def serialize(self, value: timedelta) -> int:
        """Serialize a timedelta object to an integer representing total minutes."""
        return int(value.total_seconds() // 60)  # Convert total seconds to minutes

    def deserialize(self, value: int) -> timedelta:
        """Deserialize an integer representing total minutes to a timedelta object."""
        return timedelta(minutes=value)


@dataclass
class Calendar(DataClassDictMixin):
    """Information about the calendar tasks.

    An Automower can have several tasks. If the mower supports
    work areas the property workAreaId is required to connect
    the task to an work area.
    """

    start: time = field(
        metadata=field_options(serialization_strategy=TimeSerializationStrategy())
    )
    duration: timedelta = field(
        metadata=field_options(serialization_strategy=DurationSerializationStrategy())
    )
    monday: bool
    tuesday: bool
    wednesday: bool
    thursday: bool
    friday: bool
    saturday: bool
    sunday: bool
    work_area_id: int | None = field(
        metadata=field_options(alias="workAreaId"), default=None
    )

    class Config(BaseConfig):
        """BaseConfig for Calendar."""

        serialize_by_alias = True
        omit_none = True


@dataclass
class AutomowerCalendarEvent:
    """Information about the calendar tasks.

    Internal class for creating recurrence.
    """

    start: datetime
    duration: timedelta
    uid: str
    day_set: set[DayOfWeek]


class ConvertScheduleToCalendar:
    """Convert the Husqvarna task to an AutomowerCalendarEvent."""

    def __init__(self, task: Calendar) -> None:
        """Initialize the schedule to calendar converter."""
        self.task = task
        self.now = datetime.now()  # noqa: DTZ005
        self.begin_of_current_day = self.now.replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        self.current_day = self.now.weekday()

    def next_weekday_with_schedule(self) -> datetime:
        """Find the next weekday with a schedule entry."""
        for days in range(8):
            time_to_check = self.now + timedelta(days=days)
            time_to_check_begin_of_day = time_to_check.replace(
                hour=0, minute=0, second=0, microsecond=0
            )
            day_to_check = time_to_check.weekday()
            day_to_check_as_string = WEEKDAYS[day_to_check]
            for task_field in fields(self.task):
                field_name = task_field.name
                field_value = getattr(self.task, field_name)
                if field_value is True and field_name == day_to_check_as_string:
                    end_task = (
                        time_to_check_begin_of_day
                        + timedelta(
                            hours=self.task.start.hour, minutes=self.task.start.minute
                        )
                        + self.task.duration
                    )
                    if (
                        self.begin_of_current_day == time_to_check_begin_of_day
                        and end_task < self.now
                    ):
                        break
                    return self.now + timedelta(days=days)
        return self.now

    def make_dayset(self) -> set[DayOfWeek]:
        """Generate a set of days from a task."""
        return {
            day
            for day in (
                WEEKDAYS_TO_ICAL.get(day) for day in WEEKDAYS if getattr(self.task, day)
            )
            if day is not None
        }

    def make_event(self) -> AutomowerCalendarEvent:
        """Generate a AutomowerCalendarEvent from a task."""
        dayset = self.make_dayset()
        next_wd_with_schedule = self.next_weekday_with_schedule()
        begin_of_day_with_schedule = next_wd_with_schedule.replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        return AutomowerCalendarEvent(
            start=(
                begin_of_day_with_schedule
                + timedelta(hours=self.task.start.hour, minutes=self.task.start.minute)
            ).replace(tzinfo=tz_util.MOWER_TIME_ZONE),
            duration=self.task.duration,
            uid=f"{self.task.start}_{self.task.duration}_{dayset}",
            day_set=dayset,
        )


@dataclass
class Tasks(DataClassDictMixin):
    """DataClass for Task values."""

    tasks: list[Calendar]

    @property
    def timeline(self) -> ProgramTimeline:
        """Return a timeline of all schedules."""
        return self.timeline_tz()

    def timeline_tz(self) -> ProgramTimeline:
        """Return a timeline of all schedules."""
        schedule_no: dict[int, int] = {}
        for task in self.tasks:
            key = task.work_area_id if task.work_area_id is not None else -1
            schedule_no.setdefault(key, 0)

        iters: list[Iterable[SortableItem[Timespan, ProgramEvent]]] = []

        for task in self.tasks:
            event = ConvertScheduleToCalendar(task).make_event()
            number = self.generate_schedule_no(task, schedule_no)

            freq = (
                ProgramFrequency.DAILY
                if len(event.day_set) == 7
                else ProgramFrequency.WEEKLY
            )

            iters.append(
                create_recurrence(
                    schedule_no=number,
                    work_area_id=task.work_area_id,
                    frequency=freq,
                    dtstart=event.start,
                    duration=event.duration,
                    days_of_week=event.day_set,
                )
            )

        return ProgramTimeline(MergedIterable(iters))

    def generate_schedule_no(self, task: Calendar, schedule_no: dict[int, int]) -> int:
        """Return a schedule number."""
        key = task.work_area_id if task.work_area_id is not None else -1
        schedule_no[key] += 1
        return schedule_no[key]


def make_name_string(work_area_name: str | None, number: int) -> str:
    """Return a string for the calendar summary."""
    if work_area_name is not None:
        return f"{work_area_name} schedule {number}"
    return f"Schedule {number}"



"""Models for Automower Connect API - System."""

from dataclasses import dataclass, field

from mashumaro import DataClassDictMixin, field_options


@dataclass
class System(DataClassDictMixin):
    """System information about a Automower."""

    name: str
    model: str
    serial_number: str = field(metadata=field_options(alias="serialNumber"))



"""Library for Exception using the Husqvarna Automower API."""


class HusqvarnaAutomowerError(Exception):
    """Base class for all client Errors."""


class ApiError(HusqvarnaAutomowerError):
    """Raised during problems talking to the API."""


class FeatureNotSupportedError(HusqvarnaAutomowerError):
    """Raised when the feature is not supported by the mower."""


class WorkAreasDifferentError(HusqvarnaAutomowerError):
    """Raised when the work areas for setting the calendar are different."""


class AuthError(HusqvarnaAutomowerError):
    """Raised due to auth problems talking to API."""


class InvalidSyncTokenError(HusqvarnaAutomowerError):
    """Raised when the sync token is invalid."""


class ApiBadRequestError(HusqvarnaAutomowerError):
    """Raised due sending a Rest command resulting in a bad request."""


class ApiForbiddenError(HusqvarnaAutomowerError):
    """Raised due to permission errors talking to API."""


class ApiUnauthorizedError(HusqvarnaAutomowerError):
    """Raised occasionally, mustn't harm the connection."""


class NoDataAvailableError(HusqvarnaAutomowerError):
    """Raised due updating data, when no data is available."""


class HusqvarnaTimeoutError(HusqvarnaAutomowerError):
    """Raised due connecting the websocket."""


class HusqvarnaWSServerHandshakeError(HusqvarnaAutomowerError):
    """Raised due connecting the websocket if server not available."""


class HusqvarnaWSClientError(HusqvarnaAutomowerError):
    """Raised due connecting the websocket."""



"""Models for Automower Connect API - Metadata."""

from dataclasses import dataclass, field
from datetime import UTC, datetime

from mashumaro import DataClassDictMixin, field_options


@dataclass
class Metadata(DataClassDictMixin):
    """DataClass for Metadata values."""

    connected: bool
    status_dateteime: datetime = field(
        metadata=field_options(
            deserialize=lambda x: (datetime.fromtimestamp(x / 1000, tz=UTC)),
            alias="statusTimestamp",
        ),
    )



"""Utils for Husqvarna Automower."""

import logging
from datetime import UTC, tzinfo
from functools import lru_cache

_LOGGER = logging.getLogger(__name__)

MOWER_TIME_ZONE: tzinfo = UTC


@lru_cache(maxsize=1)
def get_mower_time_zone() -> tzinfo:
    """Get the default time zone."""
    return MOWER_TIME_ZONE


def set_mower_time_zone(time_zone: tzinfo) -> None:
    """Set a default time zone to be used when none is specified.

    Async friendly.
    """
    # pylint: disable-next=global-statement
    global MOWER_TIME_ZONE  # noqa: PLW0603

    MOWER_TIME_ZONE = time_zone
    get_mower_time_zone.cache_clear()



"""Module for AbstractAuth for Husqvarna Automower."""

import asyncio
import json
import logging
from abc import ABC, abstractmethod
from http import HTTPStatus
from typing import Any

from aiohttp import (
    ClientError,
    ClientResponse,
    ClientResponseError,
    ClientSession,
    ClientWebSocketResponse,
    WSServerHandshakeError,
)

from .const import API_BASE_URL, AUTH_HEADER_FMT, WS_URL
from .exceptions import (
    ApiBadRequestError,
    ApiError,
    ApiForbiddenError,
    ApiUnauthorizedError,
    AuthError,
    HusqvarnaWSClientError,
    HusqvarnaWSServerHandshakeError,
)
from .utils import structure_token

ERROR = "error"
STATUS = "status"
MESSAGE = "message"

_LOGGER = logging.getLogger(__name__)


class AbstractAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: ClientSession, host: str) -> None:
        """Initialize the auth."""
        self._websession = websession
        self._host = host if host is not None else API_BASE_URL
        self._client_id = ""
        self.loop = asyncio.get_event_loop()
        self.ws_status: bool = True
        self.ws: ClientWebSocketResponse

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def request(self, method: str, url: str, **kwargs: Any) -> ClientResponse:
        """Make a request."""
        headers = await self.headers()
        if not url.startswith(("http://", "https://")):
            url = f"{self._host}/{url}"
        _LOGGER.debug("request[%s]=%s %s", method, url, kwargs.get("params"))
        if method != "get" and "json" in kwargs:
            _LOGGER.debug("request[%s json]=%s", method, kwargs["json"])
        return await self._websession.request(method, url, **kwargs, headers=headers)

    async def get(self, url: str, **kwargs: Any) -> ClientResponse:
        """Make a get request."""
        try:
            resp = await self.request("get", url, **kwargs)
        except ClientError as err:
            raise ApiError(err) from err
        return await AbstractAuth._raise_for_status(resp)

    async def get_json(self, url: str, **kwargs: Any) -> Any:
        """Make a get request and return json response."""
        resp = await self.get(url, **kwargs)
        try:
            result = await resp.json(encoding="UTF-8")
        except ClientError as err:
            raise ApiError(err) from err
        if not isinstance(result, dict):
            raise ApiError(result) from result
        _LOGGER.debug("response=%s", json.dumps(result))
        return result

    async def post(self, url: str, **kwargs: Any) -> ClientResponse:
        """Make a post request."""
        try:
            resp = await self.request("post", url, **kwargs)
        except ClientError as err:
            raise ApiError(err) from err
        return await AbstractAuth._raise_for_status(resp)

    async def post_json(self, url: str, **kwargs: Any) -> dict[str, Any]:
        """Make a post request and return a json response."""
        resp = await self.post(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiError(err) from err
        if not isinstance(result, dict):
            raise ApiError(result) from result
        _LOGGER.debug("response=%s", json.dumps(result))
        return result

    async def patch(self, url: str, **kwargs: Any) -> ClientResponse:
        """Make a post request."""
        try:
            resp = await self.request("patch", url, **kwargs)
        except ClientError as err:
            raise ApiError(err) from err
        return await AbstractAuth._raise_for_status(resp)

    async def patch_json(self, url: str, **kwargs: Any) -> dict[str, Any]:
        """Make a post request and return a json response."""
        resp = await self.patch(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiError(err) from err
        if not isinstance(result, dict):
            raise ApiError(result) from result
        _LOGGER.debug("response=%s", json.dumps(result))
        return result

    async def _async_get_access_token(self) -> str:
        """Request a new access token."""
        try:
            return await self.async_get_access_token()
        except ClientError as err:
            raise AuthError(err) from err

    async def headers(self) -> dict[str, str]:
        """Generate headers for ReST requests."""
        access_token = await self._async_get_access_token()
        if not self._client_id:
            token_structured = structure_token(access_token)
            self._client_id = token_structured.client_id
        return {
            "Authorization": f"Bearer {access_token}",
            "Authorization-Provider": "husqvarna",
            "Content-Type": "application/vnd.api+json",
            "X-Api-Key": self._client_id,
        }

    @staticmethod
    async def _raise_for_status(resp: ClientResponse) -> ClientResponse:
        """Raise Errors on failure methods."""
        detail = await AbstractAuth._error_detail(resp)
        try:
            resp.raise_for_status()
        except ClientResponseError as err:
            if err.status == HTTPStatus.BAD_REQUEST:
                raise ApiBadRequestError(err) from err
            if err.status == HTTPStatus.UNAUTHORIZED:
                raise ApiUnauthorizedError(err) from err
            if err.status == HTTPStatus.FORBIDDEN:
                raise ApiForbiddenError(err) from err
            detail.append(err.message)
            raise ApiError(": ".join(detail)) from err
        except ClientError as err:
            raise ApiError(err) from err
        return resp

    @staticmethod
    async def _error_detail(resp: ClientResponse) -> list[str]:
        """Return an error message string from the API response."""
        if resp.status < 400:
            return []
        try:
            result = await resp.json()
            error = result.get(ERROR, {})
        except ClientError:
            return []
        message = ["Error from API", f"{resp.status}"]
        if STATUS in error:
            message.append(f"{error[STATUS]}")
        if MESSAGE in error:
            message.append(error[MESSAGE])
        return message

    async def websocket_connect(self) -> None:
        """Start a websocket connection."""
        token = await self._async_get_access_token()
        try:
            self.ws = await self._websession.ws_connect(
                url=WS_URL,
                headers={"Authorization": AUTH_HEADER_FMT.format(token)},
                heartbeat=60,
            )
        except WSServerHandshakeError as err:
            raise HusqvarnaWSServerHandshakeError(err) from err
        except ClientError as err:
            raise HusqvarnaWSClientError(err) from err



"""Models for Automower Connect API - Statistics."""

from dataclasses import dataclass, field

from mashumaro import DataClassDictMixin, field_options


@dataclass
class Statistics(DataClassDictMixin):
    """DataClass for Statistics values."""

    cutting_blade_usage_time: int | None = field(
        metadata=field_options(alias="cuttingBladeUsageTime"), default=None
    )
    """The number of seconds since the last reset of the cutting blade usage counter."""

    downtime: int | None = field(metadata=field_options(alias="downTime"), default=None)
    """The number of seconds the mower has been disconnected from the cloud.
    Not available on all models."""

    number_of_charging_cycles: int | None = field(
        metadata=field_options(alias="numberOfChargingCycles"), default=None
    )
    """Number of charging cycles."""

    number_of_collisions: int | None = field(
        metadata=field_options(alias="numberOfCollisions"), default=None
    )
    """The total number of collisions."""

    total_charging_time: int | None = field(
        metadata=field_options(alias="totalChargingTime"), default=None
    )
    """Total charging time in seconds."""

    total_cutting_time: int | None = field(
        metadata=field_options(alias="totalCuttingTime"), default=None
    )
    """Total cutting time in seconds."""

    total_drive_distance: int | None = field(
        metadata=field_options(alias="totalDriveDistance"), default=None
    )
    """Total drive distance in meters. It's a calculated value based on totalRunningTime
     multiply with average speed for the mower depending on the model."""

    total_running_time: int | None = field(
        metadata=field_options(alias="totalRunningTime"), default=None
    )
    """The total running time in seconds. (the wheel motors have been running)"""

    total_searching_time: int | None = field(
        metadata=field_options(alias="totalSearchingTime"), default=None
    )
    """The total searching time in seconds."""

    uptime: int | None = field(metadata=field_options(alias="upTime"), default=None)
    """The number of seconds the mower has been connected to the cloud."""



"""Models for Automower Connect API - Capabilities."""

from dataclasses import dataclass, field

from mashumaro import DataClassDictMixin, field_options


@dataclass
class Capabilities(DataClassDictMixin):
    """Information about what capabilities the Automower has."""

    can_confirm_error: bool = field(metadata=field_options(alias="canConfirmError"))
    headlights: bool
    position: bool
    stay_out_zones: bool = field(metadata=field_options(alias="stayOutZones"))
    work_areas: bool = field(metadata=field_options(alias="workAreas"))



"""Models for Automower Connect API - Settings."""

from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro import DataClassDictMixin, field_options


class HeadlightModes(StrEnum):
    """Headlight modes of a lawn mower."""

    ALWAYS_ON = "always_on"
    ALWAYS_OFF = "always_off"
    EVENING_ONLY = "evening_only"
    EVENING_AND_NIGHT = "evening_and_night"


@dataclass
class Headlight(DataClassDictMixin):
    """DataClass for Headlight values."""

    mode: HeadlightModes | None = field(
        metadata=field_options(
            deserialize=lambda x: HeadlightModes(x.lower()) if x else None
        ),
        default=None,
    )


@dataclass
class Settings(DataClassDictMixin):
    """DataClass for Settings values."""

    headlight: Headlight
    cutting_height: int | None = field(
        metadata=field_options(alias="cuttingHeight"), default=None
    )



"""Models for Automower Connect API for endpoint /mowers."""

from typing import Any, TypedDict

from .model_message import Message


class SystemAttributes(TypedDict):
    """Represents system attributes of the mower."""

    name: str
    model: str
    serialNumber: int


class BatteryAttributes(TypedDict):
    """Represents battery attributes of the mower."""

    batteryPercent: int


class CapabilitiesAttributes(TypedDict):
    """Represents capabilities attributes of the mower."""

    canConfirmError: bool
    headlights: bool
    position: bool
    stayOutZones: bool
    workAreas: bool


class MowerAttributes(TypedDict):
    """Represents mower-related attributes."""

    mode: str
    activity: str
    inactiveReason: str
    state: str
    workAreaId: int
    errorCode: int
    errorCodeTimestamp: int
    isErrorConfirmable: bool


class CalendarTask(TypedDict):
    """Represents a calendar task with its schedule and work area ID."""

    start: int
    duration: int
    monday: bool
    tuesday: bool
    wednesday: bool
    thursday: bool
    friday: bool
    saturday: bool
    sunday: bool
    workAreaId: int


class PlannerOverride(TypedDict):
    """Represents override settings for the mower's planner."""

    action: str


class WorkAreaAttributes(TypedDict):
    """Represents attributes of a specific work area for the mower."""

    workAreaId: int
    name: str
    cuttingHeight: int
    enabled: bool
    progress: int | None
    lastTimeCompleted: int | None


class PositionAttributes(TypedDict):
    """Represents a position with latitude and longitude."""

    latitude: float
    longitude: float


class StatisticsAttributes(TypedDict):
    """Represents mower statistics."""

    cuttingBladeUsageTime: int
    downTime: int
    numberOfChargingCycles: int
    numberOfCollisions: int
    totalChargingTime: int
    totalCuttingTime: int
    totalDriveDistance: int
    totalRunningTime: int
    totalSearchingTime: int
    upTime: int


class StayOutZone(TypedDict):
    """Represents a stay-out zone with its id, name, and enabled status."""

    id: str
    name: str
    enabled: bool


class StayOutZonesAttributes(TypedDict):
    """Represents stay-out zones and their status."""

    dirty: bool
    zones: list[StayOutZone]


class SettingsAttributes(TypedDict):
    """Represents mower settings, including cutting height and headlight mode."""

    cuttingHeight: int
    headlight: dict[str, str]


class MowerDataAttributes(TypedDict):
    """Represents attributes of the mower data."""

    system: SystemAttributes
    battery: BatteryAttributes
    capabilities: CapabilitiesAttributes
    mower: MowerAttributes
    calendar: dict[str, list[CalendarTask]]
    planner: dict[str, Any]
    metadata: dict[str, Any]
    workAreas: list[WorkAreaAttributes]
    positions: list[PositionAttributes]
    statistics: StatisticsAttributes
    stayOutZones: StayOutZonesAttributes
    settings: SettingsAttributes
    messages: list[Message]


class MowerDataItem(TypedDict):
    """Represents a single mower data item with its type, id, and attributes."""

    type: str
    id: str
    attributes: MowerDataAttributes


class MowerDataResponse(TypedDict):
    """Represents the full response structure containing mower data."""

    data: list[MowerDataItem]



"""Models for Husqvarna Automower data."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime  # noqa:TC003
from enum import StrEnum

from mashumaro import DataClassDictMixin, field_options

from .model_mower import (
    ERRORCODES,
    snake_case,
)
from .utils import convert_timestamp_to_aware_datetime


class Severity(StrEnum):
    """Severity level of a diagnostic message."""

    FATAL = "fatal"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    DEBUG = "debug"
    SW = "sw"
    UNKNOWN = "unknown"


@dataclass
class Message(DataClassDictMixin):
    """Single diagnostic or error message."""

    time: datetime = field(
        metadata=field_options(
            deserialize=convert_timestamp_to_aware_datetime,
            alias="time",
        ),
    )
    code: str = field(
        metadata=field_options(
            deserialize=lambda x: snake_case(ERRORCODES.get(x)),
            alias="code",
        )
    )
    severity: Severity = field(
        metadata=field_options(
            alias="severity",
            deserialize=lambda x: Severity(x.lower()),
        )
    )
    latitude: float = field(metadata=field_options(alias="latitude"))
    longitude: float = field(metadata=field_options(alias="longitude"))


@dataclass
class MessageAttributes(DataClassDictMixin):
    """Message attribute container."""

    messages: list[Message] = field(
        metadata=field_options(alias="messages"), default_factory=list
    )


@dataclass
class SingleMessageAttributes(DataClassDictMixin):
    """Single message attribute container."""

    message: Message


@dataclass
class MessageData(DataClassDictMixin):
    """Top-level message data node."""

    type: str
    id: str
    attributes: MessageAttributes = field(metadata=field_options(alias="attributes"))


@dataclass
class SingleMessageData(DataClassDictMixin):
    """Top-level single message data node."""

    type: str
    id: str
    attributes: SingleMessageAttributes
